/*
 * IBM Confidential OCO Source Materials
 * 5639-D57, 5630-A36, 5630-A37, 5724-D18 (C) Copyright IBM Corp. 1992, 2004
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
 */

/*
 * This plug-in examines the cache-control headers from the application
 * server to determine the URLs to invalidate.
 * Configuration:
 *    Transmogrifier <modrw_plugin>:modrw_open:modrw_write:modrw_close:modrw_error
 * Where <cacheinv_pb> is the fully qualifed filename of this plug-in.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "HTAPI.h"

static char* MODRW_CONTENT_TYPE[] = { "text/html", "text/css", NULL};

#define TRUE 1
#define FALSE 0

#ifdef _WIN32_WINNT
#define strncasecmp strnicmp
#endif
#define ALPHA(x) ((x >= 'a' && x <= 'z') || (x >= 'A' && x <= 'Z'))
#define WHITESPACE(x) (x == ' ' || x == '\t' )

#ifndef strncasecomp
int strncasecomp(char* cp1, char*cp2, int n) {
        int i;
        for(i=0; i<n; i++) {
                if ( ((*cp1 >= 'a' && *cp1 <= 'z') ? (*cp1 + 'A' - 'a') : *cp1) !=
                     ((*cp2 >= 'a' && *cp2 <= 'z') ? (*cp2 + 'A' - 'a') : *cp2)) return 1;
        }
        return 0;
}
#endif

/* In order to fully support junction rewrite on proxy, we need also rewrite the set-cookie headers */
/* according to RFC 2109, 2965, each Set-Cookie header may contain multiple cookie values seperated */
/* by comma, and each request may contain multiple Set-Cookie headers. Following 2 function will get*/
/* Set-Cookie header from api, and compose new list of headers                                      */
char* handleSetCookie(unsigned char* handle,
                      unsigned char* prefix,
                      unsigned char* host,
                      unsigned char* incominghost,
                      unsigned char* bcp,
                      unsigned char* cookie,
                      unsigned char* end)
{unsigned char *startp, *cp;
 unsigned char orgpath[256], *orgpathp = NULL;
 unsigned char orgdomain[128];
 unsigned char *path = NULL;
 int i;
        if (cookie == NULL || *cookie == '\0') return bcp;
        orgpath[0] = orgdomain[0] = 0;
        startp = cp = (unsigned char*)cookie;
        /* search the path and domain value of this cookie */
        do {
              if (*cp == ';') cp ++;
              while (*cp == ' ' || *cp == '\t') cp++;
              if (*cp == 0) break;
              if (!strncasecomp(cp, "path", 4)) {
                        cp += 4;
                        while (*cp == ' ' || *cp == '\t') cp++;
                        if (*cp != '=') continue;
                        else cp++;
                        while (*cp == ' ' || *cp == '\t') cp++;
                        orgpathp = cp;
                        for(i=0; i<sizeof(orgpath); i++) {
                                if (cp > end || *cp == '\0' || *cp == ' ' ||
                                    *cp == '\t' || *cp == ';') break;
                                orgpath[i] = *cp;
                                cp++;
                        }
                        orgpath[i] = '\0';
              } else if (!strncasecomp(cp, "domain", 6)) {
                        cp +=6;
                        while (*cp == ' ' || *cp == '\t') cp++;
                        if (*cp != '=') continue;
                        else cp++;
                        while (*cp == ' ' || *cp == '\t') cp++;
                        for(i=0; i<sizeof(orgdomain); i++) {
                                if (cp > end || *cp == '\0' || *cp == ' ' ||
                                    *cp == '\t' || *cp == ';') break;
                                orgdomain[i] = *cp;
                                cp++;
                        }
                        orgdomain[i] = '\0';
             }
             cp = strchr(cp, ';');
        } while(cp <= end && cp != NULL);
        /* if there is a domain, verify it with host header */
        if (*orgdomain) {
            /* The cookie is not for this host, just copy it. */
            for (i=0; i<sizeof(orgdomain); i++) {
                if(orgdomain[i] >= 'a' && orgdomain[i] <= 'z') orgdomain[i] += ('A' - 'a');
                else if (orgdomain[i] == '\0') break;
            }
            if(!strstr(host, orgdomain) && (incominghost ? !strstr(incominghost, orgdomain) : 1)) {
                strncpy(bcp, cookie, (end - cookie + 1));
                bcp += (end - cookie + 1);
                return bcp;
            }
        }
 
        /* No path specified, add the path information */
        if (!orgpathp) {
                strncpy(bcp, cookie, (end - cookie + 1));
                bcp += (end - cookie + 1);
                if(*(bcp - 1) != ';') {
                        *bcp = ';';
                        bcp++;
                }
                bcp += sprintf(bcp, " Path=%s", prefix);
                return bcp;
        }
 
        /* Change the path to prefix + path */
        memcpy(bcp, cookie, orgpathp - cookie);
        bcp += orgpathp - cookie;
        memcpy(bcp, prefix, strlen(prefix));
        bcp += strlen(prefix);
        if (strlen(orgpath) == 1) orgpathp ++;
        strncpy(bcp, orgpathp, end - orgpathp + 1);
        bcp += end - orgpathp + 1;
        return bcp;
}

/* Threre maybe multiple Set-Cookie headers, we need iterate them all */
int rewriteSetCookieHeaders(unsigned char* handle, unsigned char* prefix)
{unsigned long n = 0, nn = 0;
 unsigned char* buffer[21];
 const unsigned char* cookie = NULL;
 unsigned char backup, *host, *incominghost = NULL;
 unsigned char *startp, *endp, *cp, *bcp, *path;
 int prefixlen, i;
 unsigned long addHeader = 0;
 unsigned char* reqhdr;
 
        for(i=0; i<21; i++) {
                buffer[i] = NULL;
        }
 
        prefixlen = strlen(prefix);
        cookie = (unsigned char*)httpd_getvar(handle, "PROXY_SET-COOKIE", &n);
        host   = (unsigned char*)httpd_getvar(handle, "HTTP_HOST", &n);
        reqhdr = (unsigned char*)httpd_getvar(handle, "REQHDR", &n);
        if (reqhdr) {
            for (i=0; i<strlen(reqhdr); i++) {
                if(reqhdr[i] >= 'a' && reqhdr[i] <= 'z') reqhdr[i] += ('A' - 'a');
                else if (reqhdr[i] == '\0') break;
            }
            cp = strstr(reqhdr, "HOST:");
            if (cp) {
                cp += 5;
                while(*cp== ' ' || *cp == '\t') cp++;
                incominghost = cp;
                while(*cp != ' ' && *cp != '\t' && *cp != '\0' && *cp != '\r' && *cp != '\n') cp++;
                *cp = 0;
            }
        }
 
        if (!cookie || *cookie == '0' || !host || *host == '\0') return 0;
 
        /* Change the host to capital letter */
        for (i=0; i<strlen(host); i++) {
            if(host[i] >= 'a' && host[i] <= 'z') host[i] += ('A' - 'a');
            else if (host[i] == '\0') break;
        }
 
        /* iterate each Set-Cookie header and rewrite them */
        while(cookie) {
 
                /* Allocate 4K buffer for set-cookie header */
                buffer[n] = (char*)malloc(4208);
                if (!buffer[n]) break;
                else bcp = buffer[n];
                startp = cp = (unsigned char*)cookie;
                endp   = startp + strlen(cookie);
                while(cp) {
NEXT:

                        cp = strchr(cp, ',');
                        if (cp) {
                            /* In case of old styled cookie with expire property */
                            unsigned char *ccp;
                            ccp = cp;
                            while (*ccp != '=' && ccp > startp) ccp--;
                            if (ccp > startp) {
                                while(*ccp !=  ' ' && *ccp != '\t' && *ccp !=  ',' && *ccp !=  ';' && ccp > startp) ccp--;
                                ccp++;
                                if (!strncasecomp(ccp, "expires", 7)) {
                                    do {
                                        cp++;
                                    } while(*cp != ';' && *cp !=',' && cp < endp);
                                    if (*cp != ',') goto NEXT;
                                }
                            }
                        }
 
                        if (!cp) break;
                        if (cp - startp + prefixlen + 10 > 4208) {
                                *bcp = 0;
                                bcp = NULL;
                                n++;
                                if (n >= 19) break;
                                buffer[n] = (char*)malloc(4208);
                                if (!buffer[n]) break;
                                else bcp = buffer[n];
                        }
                        backup = *cp; *cp = '\0';
                        bcp = handleSetCookie(handle, prefix, host, incominghost, bcp, startp, cp - 1);
                        *cp = backup;
                        *bcp = *cp;
                        bcp++; cp++;
                        startp = cp;
                }
                if (bcp) {
                    if (startp) {
                        bcp = handleSetCookie(handle, prefix, host, incominghost, bcp, startp, endp - 1);
                    }
                    *bcp = 0;
                }
                addHeader = HTTPD_SETVAR_REMOVE;
                httpd_setvar(handle, "HTTP_SET-COOKIE", buffer[n], &addHeader);
                if (n >= 19) break;
                else n++;
                cookie = httpd_getvar(handle, "PROXY_SET-COOKIE", &nn);
        }
        addHeader = HTTPD_SETVAR_ADD;
        for(i=0; i<21; i++) {
                if (i<n) httpd_setvar(handle, "HTTP_SET-COOKIE", buffer[i], &addHeader);
                if (buffer[i]) free(buffer[i]);
        }
        return n;
}


void *
HTTPD_LINKAGE
modrw_open (unsigned char *handle,
	    long          *retc)
{
    unsigned long n = 0;
    int ix, do_rewrite = FALSE;
    char *valid_ctypeP;
    unsigned char *modrw_prefix = NULL;
    unsigned char *modrw_content_type = NULL;
    unsigned char trace_string[1024];
    long tracestrlen, rc;
	

    *retc = HTTP_NOACTION;
    HTTPD_junctionrewritesetcookie(handle, &rc);
    modrw_prefix = httpd_getvar(handle, "_THP_MODRW_PREFIX", &n);
    if (NULL != modrw_prefix) {
	modrw_content_type = httpd_getvar(handle, "PROXY_CONTENT_TYPE", &n);
	if (modrw_content_type != NULL) {
	    /* clean up string */
	    while (modrw_content_type && WHITESPACE(*modrw_content_type)) modrw_content_type++;
	    for (ix = 0, valid_ctypeP = MODRW_CONTENT_TYPE[0]; valid_ctypeP; ix++, valid_ctypeP = MODRW_CONTENT_TYPE[ix]) {
		if (strncasecmp(modrw_content_type, valid_ctypeP, strlen(valid_ctypeP)) == 0) {
		    do_rewrite = TRUE;
		    break;
		}
	    }
	    /* Prefix there but response is not rewritable */
	    sprintf(trace_string, "modrw: %s URL for this content (%s)\n",
		    do_rewrite == TRUE ? "Rewriting" : "Will not rewrite", modrw_content_type);
	    tracestrlen = strlen(trace_string);
	    HTTPD_log_trace(handle, trace_string, &tracestrlen, &rc);
	    if (do_rewrite) {
		n = HTTPD_SETVAR_REPLACE;
		rc = httpd_setvarint(handle, "_THP_MODRW_ENABLE", TRUE);
                HTTPD_open(handle, &rc);
		*retc = HTTP_OK;
	    }
	}
    }
    return(NULL);
}

void
HTTPD_LINKAGE
modrw_write (unsigned char *handle,
        unsigned char *data,
        unsigned long *length,
        void          *correlator,
        long          *retc)
{
    unsigned long n = 0;
    unsigned int modrw_enabled = 0;
    unsigned char *modrw_prefix = NULL;
    unsigned char *rwBuf = NULL;
    unsigned long rwBufLen;
    
    unsigned char trace_string[1024];
    long tracestrlen, rc = 0, rc1;

    rc = httpd_getvarint(handle, "_THP_MODRW_ENABLE", &modrw_enabled);
    if (rc == HTTPD_SUCCESS && modrw_enabled ) {
	if (modrw_prefix = httpd_getvar(handle, "_THP_MODRW_PREFIX", &n)) {
	    /* Call rewrite API */
	    rwBufLen = 0;
	    HTTPD_urlrewrite(handle, data, length, modrw_prefix, &rwBuf, &rwBufLen, &rc);
	    sprintf(trace_string, "Rewriting URL in buffer with prefix %s / rewritten buffer (%d) return code (%d)\n",
		    modrw_prefix, rwBufLen, rc);
	    tracestrlen = strlen(trace_string);
	    HTTPD_log_trace(handle, trace_string, &tracestrlen, &rc1);
	    if (rc != HTTPD_SUCCESS) {
		/* Error handling */
		/* Get and write unprocessed tags from the last pass */
		if (rwBuf) free(rwBuf);
		if (rwBuf = httpd_getvar(handle, "_THP_MODRW_TAGBUF", &n)) {
		    rwBufLen = 0;
		    rc = httpd_getvarint(handle, "_THP_MODRW_TAGLEN", &rwBufLen);
		    if (rwBufLen) {
			HTTPD_write(handle, rwBuf, &rwBufLen, &rc);
			sprintf(trace_string, "modrw: Writedata (rc:%d)\n", rc);
			tracestrlen = strlen(trace_string);
			HTTPD_log_trace(handle, trace_string, &tracestrlen, &rc);
		    }
		}
		rwBufLen = *length;
		rwBuf = data;
		rc = httpd_setvarint(handle, "_THP_MODRW_ENABLE", FALSE);
	    }
	} else {
	    rwBufLen = *length;
	    rwBuf = data;
	}
    } else {
	rwBufLen = *length;
	rwBuf = data;
    }

    
    /* just pass the data through */
    HTTPD_write(handle, rwBuf, &rwBufLen, &rc);
    if (rwBuf != NULL && rwBuf != data) free(rwBuf);
    sprintf(trace_string, "modrw: Writedata (rc:%d)\n", rc);
    tracestrlen = strlen(trace_string);
    HTTPD_log_trace(handle, trace_string, &tracestrlen, &rc);

    *retc = HTTP_OK;
    return;
}

void
HTTPD_LINKAGE
modrw_close (unsigned char *handle,
        void          *correlator,
        long          *retc)
{

    HTTPD_close(handle, retc);

    *retc = HTTP_OK;
    return;
}

void
HTTPD_LINKAGE
modrw_error (unsigned char *handle,
        void          *correlator,
        long          *retc)
{

    *retc = HTTP_OK;
    return;
}

/* default global targets init for mod rewrite */
//MJM-20020703 - Added major_version and minor_version params below.
void
HTTPD_LINKAGE
modrw_init(unsigned char *handle,
           unsigned long *major_version,
           unsigned long *minor_version,
            long *return_code) {
    HTTPD_addrwtag(handle, return_code, "!--",    "url", NULL);
    HTTPD_addrwtag(handle, return_code, "a",      "href", NULL);
    HTTPD_addrwtag(handle, return_code, "applet", "archive", "codebase", NULL);
    HTTPD_addrwtag(handle, return_code, "area",   "href", NULL);
    HTTPD_addrwtag(handle, return_code, "base",   "href", NULL);
    HTTPD_addrwtag(handle, return_code, "blockquote", "cite", NULL);
    HTTPD_addrwtag(handle, return_code, "body",   "background", NULL);
    HTTPD_addrwtag(handle, return_code, "del",    "cite", NULL);
    HTTPD_addrwtag(handle, return_code, "embed",  "pluginspage", NULL);
    HTTPD_addrwtag(handle, return_code, "form",   "action", NULL);
    HTTPD_addrwtag(handle, return_code, "frame",  "src", "longdesc", NULL);
    HTTPD_addrwtag(handle, return_code, "head",   "profile", NULL);
    HTTPD_addrwtag(handle, return_code, "input",  "src", "usemap", NULL);
    HTTPD_addrwtag(handle, return_code, "iframe", "src", "longdesc", NULL);
    HTTPD_addrwtag(handle, return_code, "ilayer", "src", "background", NULL);
    HTTPD_addrwtag(handle, return_code, "img",    "src", "usemap", "lowsrc", "longdesc", "dynsrc", NULL);
    HTTPD_addrwtag(handle, return_code, "ins",    "cite", NULL);
    HTTPD_addrwtag(handle, return_code, "layer",  "src", "background", NULL);
    HTTPD_addrwtag(handle, return_code, "link",   "href", NULL);
    HTTPD_addrwtag(handle, return_code, "meta",   "url", NULL);
    HTTPD_addrwtag(handle, return_code, "object", "data", "classid^java:^javascript:", "codebase", "codepage", "usemap", NULL);
    HTTPD_addrwtag(handle, return_code, "q",      "cite", NULL);
    HTTPD_addrwtag(handle, return_code, "script", "src", NULL);
    HTTPD_addrwtag(handle, return_code, "table",  "background", NULL);
    HTTPD_addrwtag(handle, return_code, "td",     "background", NULL);
    HTTPD_addrwtag(handle, return_code, "th",     "background", NULL);
    HTTPD_addrwtag(handle, return_code, "tr",     "background", NULL);
    HTTPD_addrwtag(handle, return_code, "SPECIAL","background-image", "url", NULL);

    // MJM-20020703
    *return_code = HTTP_OK;
}

